
//		CmdTimer.cpp - Automatic command timer for TCC


#define UNICODE								1
#define _UNICODE							1

#include <stdio.h>
#include <tchar.h>

#include <windows.h>

#include "PlugIn.h"
#include "TakeCmd.h"


#define PLUGIN_NAME							L"CmdTimer"

#define VER_MAJOR							1
#define VER_MINOR							0
#define VER_BUILD							2
#define VER_PATCH							3

#ifdef _WIN64
#define PLATFORM							L"x64"
#else
#define PLATFORM							L"x86"
#endif


#define MAX_ARG_LEN							8192
#define MAX_ARGS							1024
#define SHORT_BUF_LEN						1024

#define ERRHANDLE							GetStdHandle( STD_ERROR_HANDLE )

#define LF									0x000a
#define CR									0x000d
#define QUOTE								0x0022
#define ESC									0x001b


#define SLOWCMD_DEFAULT						10
#define TICKS_PER_SECOND					10000000

#define SHOW_NOTHING						0x0000
#define SHOW_STATUSBAR						0x0001
#define SHOW_TEXTBAR_ALWAYS					0x0002
#define SHOW_BOTH							0x0003
#define SHOW_TEXTBAR_IF_STATUSBAR_FAILS		0x0004
#define SHOW_DEFAULT						0x0005

#define TEXTBAR_SAVE_SIZE					200
#define TEXTBAR_DEFAULT_COLORS				0x001f


LPTSTR arg[MAX_ARGS];
int arglen[MAX_ARGS];
int numargs = 0;
wchar_t allargs[MAX_ARG_LEN] = L"";


BOOL Running								= FALSE;
BOOL ResetTotalTime							= FALSE;
unsigned __int64 StartTime					= 0;
unsigned __int64 StopTime					= 0;
unsigned __int64 TimerDuration				= 0;
unsigned __int64 TotalDuration				= 0;
unsigned __int64 CommandCount				= 0;

DWORD ShowTimer								= SHOW_DEFAULT;
DWORD SlowCommandTime						= SLOWCMD_DEFAULT;

CHAR_INFO TextbarSaveBuffer[TEXTBAR_SAVE_SIZE];
unsigned int TextbarSaveCount				= 0;
unsigned int TextbarSaveX					= 0;
unsigned int TextbarSaveY					= 0;
WORD TextBarColors							= TEXTBAR_DEFAULT_COLORS;
const COORD TextbarSaveBufferSize			= { TEXTBAR_SAVE_SIZE, 1 };
const COORD TextbarSaveBufferCoord			= { 0, 0 };

HANDLE ConsoleBufferHandle					= NULL;


LPTSTR LastCommand							= NULL;
LPTSTR SlowCommand							= NULL;




const wchar_t PluginFeatures[] = L"cmdtimer,_cmdcount,_cmdticks,_cmdtime,_cmdtotalticks,_cmdtotaltime,_lastcmd,_ticks,@tickstotime,@tickstotimed,*key";


const wchar_t HelpWindowTitle[] = L"CmdTimer Plugin Help";


const wchar_t CmdTimerHelpText[] = L"Display or change CmdTimer plugin settings. \n\
\n\
CMDTIMER /B:n /C:n /D /K /R /S /T:n /Z slow-command \r\n\
\n\
\t/B:\tshow timer: Default Statusbar Text Both Off\n\
\t/C:\tset timer text colors\n\
\t/D \trestore plugin defaults\n\
\t/K \treset the command counter to zero\n\
\t/R \trestore settings from the registry\n\
\t/S \tsave current settings to the registry \n\
\t/T:\ttime in seconds for a slow command\n\
\t/Z \treset the total time to zero\n\
\n\
A \u2018command\u2019 is any line you type at the command line.  It might\n\
be several commands, a batch file, or whatever.  The timer starts\n\
when you press Enter, and stops when the prompt is displayed again.  \n\
If a command takes a long time, slow-command will be executed\n\
automatically when it finishes.  Use /T: to define the duration of\n\
a \u2018slow\u2019 command; 10 is the default, 0 disables this feature.\n";


const wchar_t _CmdCountHelpText[] = L"_CMDCOUNT \r\n\n\
Returns the number of times the command  \n\
counter has been started.\n";


const wchar_t _CmdTicksHelpText[] = L"_CMDTICKS \r\n\n\
Returns the duration of the last command,  \n\
in 100 ns ticks.\n";


const wchar_t _CmdTimeHelpText[] = L"_CMDTIME \r\n\n\
Returns the duration of the last command,  \n\
as H:MM:SS.mmm\n";


const wchar_t _CmdTotalTicksHelpText[] = L"_CMDTOTALTICKS \r\n\n\
Returns the total of command durations,  \n\
in 100 ns ticks.\n";


const wchar_t _CmdTotalTimeHelpText[] = L"_CMDTOTALTIME \r\n\n\
Returns the total of command durations, as  \n\
H:MM:SS.mmm\n";


const wchar_t _LastCmdHelpText[] = L"_LASTCMD \r\n\n\
Returns the first word from the last command  \n\
line entered.\n";


const wchar_t _TicksHelpText[] = L"_TICKS \r\n\n\
Returns the number of 100 ns ticks since  \n\
1601-01-01 00:00:00 UTC.\n";


const wchar_t f_TicksToTimeHelpText[] = L"@TICKSTOTIME \r\n\n\
Converts a count of 100 ns ticks to  \n\
H:MM:SS.mmm format.\n";


const wchar_t f_TicksToTimeDHelpText[] = L"@TICKSTOTIMED \r\n\n\
Converts a count of 100 ns ticks to  \n\
DDD HH:MM:SS.mmm format.\n";



const wchar_t PluginRegKey[]			= L"Software\\JPPlugins\\CmdTimer";
const wchar_t ShowTimerRegValue[]		= L"ShowTimer";
const wchar_t SlowTimeRegValue[]		= L"SlowCommandTime";
const wchar_t SlowCommandRegValue[]		= L"SlowCommand";
const wchar_t TextColorsRegValue[]		= L"TextBarColors";

const wchar_t TimerFormatEnvVar[]		= L"CMDTIMERFMT";


const wchar_t ColorName[16][15] = { L"black", L"blue", L"green", L"cyan", L"red", L"magenta", L"yellow", L"white",
	L"bright black", L"bright blue", L"bright green", L"bright cyan", L"bright red", L"bright magenta", L"bright yellow", L"bright white" };


unsigned int HotkeysTccVer		= 0;
BOOL NewKeyHook					= FALSE;


typedef struct {
	int	nKey;				// key entered
	int	nHomeRow;			// start row
	int	nHomeColumn;		// start column
	int	nRow;				// current row in window
	int	nColumn;			// current column in window
	LPTSTR	pszLine;		// command line
	LPTSTR	pszCurrent;		// pointer to position in line
	int	fRedraw;			// if != 0, redraw the line
	LPTSTR pszKey;			// (v24+ only) ASCII name of key (for example, "Ctrl-Alt-Home")
} KEYINFO24, *LPKEYINFO24;



int InitHotkeys()
{
	wchar_t temp[SHORT_BUF_LEN]			= L"%_4ver";
	unsigned int Version				= 0;


	ExpandVariables( temp, 1 );

	_set_errno( 0 );
	Version = wcstoul( temp, NULL, 10 );

	HotkeysTccVer = 0;

	if ( !errno ) {
		if ( ( Version >= 8 ) && ( Version <= 100 ) )
			HotkeysTccVer = Version;
	}

	NewKeyHook = ( HotkeysTccVer >= 24 );
	return HotkeysTccVer;
}



int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%_batch";

	Qprintf( ERRHANDLE, L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_ARG_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables( shortbuf, 1 );
	Qprintf( GetStdHandle(STD_ERROR_HANDLE), shortbuf );

	return 0;
}



BOOL PipeShell()
{
	wchar_t temp[SHORT_BUF_LEN]			= L"%[_transient]%[_pipe]";


	ExpandVariables( temp, 1 );

	if ( wcscmp( temp, L"00" ) )
		return TRUE;
	else
		return FALSE;
}



int TrimLine( LPTSTR String )
{
	if ( !String )
		return -1;
	else if ( String[0] == '\0' )
		return 0;

	int FirstNonSpace = -1, LastNonSpace = -1;
	wchar_t ch = '\0';

	for ( int i = 0; String[i]; i++ ) {
		ch = String[i];

		if ( !iswspace( ch ) ) {
			LastNonSpace = i;

			if ( FirstNonSpace == -1 )
				FirstNonSpace = i;
		}
	}

	if ( FirstNonSpace == -1 ) {
		String[0] = '\0';
		return 0;
	}

	String[LastNonSpace + 1] = '\0';

	if ( FirstNonSpace ) {
		int i = 0;

		do {
			ch = String[FirstNonSpace + i];
			String[i++] = ch;

		} while ( ch );
	}

	return 0;
}



int ParseInt( LPTSTR inString, unsigned int *outValue, wchar_t **EndChar )
{
	if ( !inString || !outValue )
		return -1;

	unsigned int value = 0;
	wchar_t *sep1 = NULL;

	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		value = wcstoul( inString + 2, &sep1, 16 );

		if ( errno )
			return -1;
		else if ( !EndChar && ( *sep1 != '\0' ) )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar = sep1;

		return 0;
	}
	else if ( iswdigit( inString[0] ) ) {
		//  parse decimal

		value = wcstoul( inString, &sep1, 10 );

		if ( errno )
			return -1;
		else if ( !EndChar && ( *sep1 != '\0' ) )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}

	return -1;
}



int ParseInt64( LPTSTR inString, unsigned __int64 *outValue )
{
	if ( !inString || !outValue )
		return -1;

	unsigned __int64 value = 0;
	wchar_t *sep1 = NULL;

	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		value = _wcstoui64( inString + 2, &sep1, 16 );

		if ( errno )
			return -1;
		else if ( *sep1 != '\0' )
			return -1;

		*outValue = value;
		return 0;
	}
	else if ( iswdigit( inString[0] ) ) {
		//  parse decimal

		value = _wcstoui64( inString, &sep1, 10 );

		if ( errno )
			return -1;
		else if ( *sep1 != '\0' )
			return -1;

		*outValue = value;
		return 0;
	}

	return -1;
}



int SaveOptions( HKEY RootKey )
{
	HKEY KeyHandle	= NULL;
	DWORD TempDW	= 0;
	int rv			= 0;


	if ( RegCreateKeyEx( RootKey, PluginRegKey, 0, NULL, 0, KEY_READ | KEY_WRITE, NULL, &KeyHandle, NULL ) == ERROR_SUCCESS ) {

		if ( RegSetValueEx( KeyHandle, ShowTimerRegValue, 0, REG_DWORD, (BYTE*) &ShowTimer, sizeof( DWORD ) ) != ERROR_SUCCESS )
			rv = 2;

		if ( RegSetValueEx( KeyHandle, SlowTimeRegValue, 0, REG_DWORD, (BYTE*) &SlowCommandTime, sizeof( DWORD ) ) != ERROR_SUCCESS )
			rv = 2;

		if ( SlowCommand ) {
			if ( RegSetValueEx( KeyHandle, SlowCommandRegValue, 0, REG_SZ, (BYTE*) SlowCommand, ( (int) wcslen( SlowCommand ) + 1 ) * sizeof( wchar_t ) ) != ERROR_SUCCESS )
				rv = 2;
		}
		else {
			wchar_t null = '\0';

			if ( RegSetValueEx( KeyHandle, SlowCommandRegValue, 0, REG_SZ, (BYTE*) &null, sizeof( wchar_t ) ) != ERROR_SUCCESS )
				rv = 2;
		}

		TempDW = TextBarColors;
		if ( RegSetValueEx( KeyHandle, TextColorsRegValue, 0, REG_DWORD, (BYTE*) &TempDW, sizeof( DWORD ) ) != ERROR_SUCCESS )
			rv = 2;

		RegCloseKey( KeyHandle );
	}
	else
		rv = 1;

	return rv;
}



LPCTSTR RootKeyName( int n )
{
	static const wchar_t RootName[2][5]		= { L"HKCU", L"HKLM" };


	if ( n == 1 )
		return RootName[0];
	else if ( n == 2 )
		return RootName[1];
	else
		return NULL;
}



int SetSlowCommand( LPTSTR Text )
{
	if ( SlowCommand ) {
		free( SlowCommand );
		SlowCommand = NULL;
	}


	if ( Text ) {
		if ( Text[0] ) {
			size_t len = wcslen( Text ) + 1;
			SlowCommand = (LPTSTR) malloc( len * sizeof( wchar_t ) );

			if ( !SlowCommand )
				return 1;

			wcscpy_s( SlowCommand, len, Text );
		}
	}

	return 0;
}



int RestoreOptions( HKEY MainKey )
{
	HKEY KeyHandle	= NULL;
	DWORD buflen	= 0;
	DWORD TempDW	= 0xffff0000;
	wchar_t TempString[MAX_ARG_LEN]	= L"";
	DWORD type		= 0;
	int rv = 1, rqrv = 0;


	if ( RegOpenKeyEx( MainKey, PluginRegKey, 0, KEY_READ, &KeyHandle ) != ERROR_SUCCESS )
		return 1;


	buflen = sizeof( DWORD );
	TempDW = 0xffff0000;
	type   = 0;
	rqrv   = RegQueryValueEx( KeyHandle, ShowTimerRegValue, 0, &type, (BYTE*) &TempDW, &buflen );

	if ( ( rqrv == ERROR_SUCCESS ) && ( type == REG_DWORD ) && ( ( TempDW & 0xffff0000 ) == 0 ) ) {
		ShowTimer = TempDW;
		rv = 0;
	}


	buflen = sizeof( DWORD );
	TempDW = 0;
	type   = 0;
	rqrv   = RegQueryValueEx( KeyHandle, SlowTimeRegValue, 0, &type, (BYTE*) &TempDW, &buflen );

	if ( ( rqrv == ERROR_SUCCESS ) && ( type == REG_DWORD ) ) {
		SlowCommandTime = TempDW;
		rv = 0;
	}


	buflen = MAX_ARG_LEN * sizeof( wchar_t );
	type   = 0;
	rqrv   = RegQueryValueEx( KeyHandle, SlowCommandRegValue, 0, &type, (BYTE*) TempString, &buflen );

	if ( ( rqrv == ERROR_SUCCESS ) && ( type == REG_SZ ) )
		SetSlowCommand( TempString );
	
	
	buflen = sizeof( DWORD );
	TempDW = 0xffff0000;
	type   = 0;
	rqrv   = RegQueryValueEx( KeyHandle, TextColorsRegValue, 0, &type, (BYTE*) &TempDW, &buflen );

	if ( ( rqrv == ERROR_SUCCESS ) && ( type == REG_DWORD ) && ( ( TempDW & 0xffffff00 ) == 0 ) && ( ( ( TempDW & 0x00f0 ) >> 4 ) != ( TempDW & 0x000f ) ) )
		TextBarColors = (WORD) TempDW;


	RegCloseKey( KeyHandle );
	return rv;
}



int RestoreOptions()
{
	if ( !RestoreOptions( HKEY_CURRENT_USER ) )
		return 1;

	if ( !RestoreOptions( HKEY_LOCAL_MACHINE ) )
		return 2;

	return 0;
}



int RestorePluginDefaults()
{
	ShowTimer			= SHOW_DEFAULT;
	SlowCommandTime		= SLOWCMD_DEFAULT;
	TextBarColors		= TEXTBAR_DEFAULT_COLORS;
	SetSlowCommand( NULL );
	return 0;
}



LPCTSTR ShowModeStr( DWORD Mode )
{
	static const wchar_t ModeStr[7][29]		= { L"Off", L"Status bar", L"Text", L"Both status bar and text", L"Default", L"Default (status bar or text)", L"Unknown" };


	if ( Mode > 6 )
		return ModeStr[6];
	else
		return ModeStr[Mode];
}



BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}



DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/cmdtimer.html";
	piInfo.pszDescription	= (LPTSTR) L"Automatic command timer for Take Command";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

	return &piInfo;
}



DLLExports BOOL WINAPI InitializePlugin( void )
{
	InitHotkeys();

	ConsoleBufferHandle = CreateFile( L"CONOUT$", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );

	if ( ( ConsoleBufferHandle == INVALID_HANDLE_VALUE ) || !ConsoleBufferHandle ) {
		if ( !PipeShell() ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error opening console buffer!\r\n" );
			Qprintf( ERRHANDLE, L"%s plugin not loaded.\r\n", PLUGIN_NAME );
		}

		return -1;
	}


	RestorePluginDefaults();
	RestoreOptions();

	wchar_t shortbuf[SHORT_BUF_LEN] = L"%[_transient]%[_pipe]%[noloadmsg]";
	ExpandVariables( shortbuf, 1 );

	if ( !wcscmp( shortbuf, L"00" ) )
		Printf( L"%s v%d.%d.%d loaded.\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );

	return 0;
}



DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{
	CloseHandle( ConsoleBufferHandle );

	free( LastCommand );
	free( SlowCommand );
	return 0;
}





#define PARSE_BREAK_SPACES		1
#define PARSE_BREAK_COMMAS		2
#define PARSE_SLASHES_KLUDGE	4
#define	PARSE_QUOTES_KLUDGE		8
#define PARSE_EQUALS_KLUDGE		16
#define PARSE_ONE_ARG_KLUDGE	32
#define PARSE_RETAIN_QUOTES		64
#define PARSE_FORCE_UPPERCASE	128
#define PARSE_NO_TRIM			256
#define PARSE_IGNORE_QUOTES		512
#define PARSE_EQ_QT_KLUDGE		1024


int ParseArgs( LPTSTR inString, unsigned int flags )
{
	/*
		flags:  bit 0 - break args at spaces
				bit 1 - break args at commas (either bit 0 or bit 1 must be set)
				bit 2 - slashes kludge:   treat /a/b like /a /b
				bit 3 - quotes kludge:    treat /a"foo" like /a:"foo"
				bit 4 - equals kludge:    break at first unquoted equals sign
				bit 5 - one-arg kludge:   if arg does not begin with /, allow unquoted spaces
				bit 6 - don't swallow double quotes
				bit 7 - force uppercase
				bit 8 - don't trim spaces from the right
				bit 9 - disable all special handling of double quotes
	*/

	if ( !inString )
		return -1;

	numargs = 0;

	int argnum = 0;
	int i = 0;
	int a = 0;
	int b = 0;
	wchar_t ch = '\0';
	BOOL quote  = FALSE;
	BOOL adv    = FALSE;
	BOOL done   = FALSE;
	int eqkl    = 0;

	if ( ( flags & ( PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS | PARSE_EQUALS_KLUDGE ) ) == 0 )
		flags |= PARSE_BREAK_SPACES;

	for ( i = 0; i < MAX_ARGS; i++ ) {
		arg[i] = NULL;
		arglen[i] = 0;
	}

	for ( i = 0; !done; i++ ) {
		ch = inString[i];
		adv = FALSE;

		if ( ( ch == '"' ) && ( ( flags & PARSE_IGNORE_QUOTES ) == 0 ) ) {
			quote = !quote;

			if ( quote && ( a == 2 ) && ( arg[argnum][0] == '/' ) && ( iswalpha( arg[argnum][1] ) ) && ( flags & PARSE_QUOTES_KLUDGE ) ) {
				if ( flags & PARSE_RETAIN_QUOTES ) {
					if ( b < ( MAX_ARG_LEN - 1 ) )
						allargs[b++] = ':';
				}
				else
					ch = ':';
			}
			else if ( ( flags & PARSE_RETAIN_QUOTES ) == 0 )
				continue;
		}

		if ( iswspace( ch ) && ( a == 0 ) && !quote )
			continue;

		if ( ( a == 0 ) && ( ch != '/' ) && ( ( flags & ( PARSE_BREAK_SPACES | PARSE_ONE_ARG_KLUDGE ) ) == ( PARSE_BREAK_SPACES | PARSE_ONE_ARG_KLUDGE ) ) )
			flags = ( flags & ~PARSE_BREAK_SPACES ) | PARSE_RETAIN_QUOTES;

		if ( iswspace( ch ) && !quote && ( a == 1 ) && ( argnum > 0 ) && ( flags & PARSE_EQUALS_KLUDGE ) && ( arg[argnum][0] == '=' ) )
			continue;

		if ( iswspace( ch ) && !quote && ( flags & PARSE_BREAK_SPACES ) )
			adv = TRUE;

		if ( ( ch == ',' ) && !quote && ( flags & PARSE_BREAK_COMMAS ) )
			adv = TRUE;

		if ( ( ch == '/' ) && !quote && ( a > 1 ) && ( arg[argnum][0] == '/' ) && ( flags & PARSE_SLASHES_KLUDGE ) ) {
			argnum++;
			a = 0;
			b++;
			allargs[b] = '\0';
		}

		if ( ( ch == '=' ) && !quote && ( eqkl == 0 ) && ( a > 0 ) && ( arg[argnum][0] != '/' ) && ( flags & PARSE_EQUALS_KLUDGE ) ) {
			argnum++;
			a = 0;
			b++;
			allargs[b] = '\0';
			eqkl = argnum;
			flags &= ~( PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS );
			//- Printf( L"   * equals kludge invoked; eqkl = %d\n", eqkl );

			if ( flags & PARSE_EQ_QT_KLUDGE )
				flags |= PARSE_IGNORE_QUOTES;
		}
		else if ( ( ch == '=' ) && !quote && ( flags & PARSE_EQUALS_KLUDGE ) ) {
			eqkl = -1;
			flags &= ~( PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS );

			if ( flags & PARSE_EQ_QT_KLUDGE )
				flags |= PARSE_IGNORE_QUOTES;
		}

		if ( ch == '\0' ) {
			if ( a )
				argnum++;

			done = TRUE;
			continue;
		}

		if ( adv ) {
			argnum++;
			a = 0;
			b++;
			allargs[b] = '\0';
			quote = FALSE;
			continue;
		}

		if ( b >= ( MAX_ARG_LEN - 1 ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Buffer overflow: b = %d, MAX_ARG_LEN = %d \r\n", b, MAX_ARG_LEN );
			return -1;
		}

		if ( argnum < MAX_ARGS ) {
			if ( a == 0 )
				arg[argnum] = allargs + b;

			if ( flags & PARSE_FORCE_UPPERCASE )
				ch = towupper( ch );

			a++;
			arglen[argnum]  = a;
			allargs[b++] = ch;
			allargs[b]   = '\0';
		}

	}

	if ( ( flags & PARSE_NO_TRIM ) == 0 ) {
		for ( i = 0; i < argnum; i++ ) {
			while ( ( arglen[i] > 0 ) && iswspace( arg[i][arglen[i] - 1] ) ) {
				arglen[i]--;
				arg[i][arglen[i]] = '\0';
			}
		}
	}

	numargs = argnum;
	return argnum;
}



BOOL IsColonOrEquals( wchar_t ch )
{
	return ( ( ch == ':' ) || ( ch == '=' ) );
}



LPCTSTR Plural( unsigned int n )
{
	static const wchar_t pl[2]	= L"s";

	if ( n == 1 )
		return pl + 1;
	else
		return pl;
}



int StatusBar( LPTSTR Text )
{
	return TakeCommandIPC( L"STATUSBAR", Text );
}



int CloseTextBar()
{
	if ( !TextbarSaveCount )
		return 0;


	SMALL_RECT WriteRegion;


	WriteRegion.Top		= TextbarSaveY;
	WriteRegion.Bottom	= TextbarSaveY;
	WriteRegion.Left	= TextbarSaveX;
	WriteRegion.Right	= TextbarSaveX + TextbarSaveCount;
	WriteConsoleOutput( ConsoleBufferHandle, TextbarSaveBuffer, TextbarSaveBufferSize, TextbarSaveBufferCoord, &WriteRegion );
	TextbarSaveCount = 0;
	return 0;
}



int ShowTextBar( LPTSTR Text )
{
	CloseTextBar();

	if ( !Text )
		return 0;
	else if ( Text[0] == '\0' )
		return 0;


	CONSOLE_SCREEN_BUFFER_INFO BufferInfo;

	unsigned int len		= (unsigned int) wcslen( Text );
	unsigned int width		= 0;
	unsigned int toprow		= 0;
	unsigned int startcol	= 0;

	COORD CursorPos;
	WORD OldColors			= 0;
	DWORD CharsWritten		= 0;
	SMALL_RECT ReadRegion;
	wchar_t svch			= '\0';

	
	GetConsoleScreenBufferInfo( ConsoleBufferHandle, &BufferInfo );

	OldColors = BufferInfo.wAttributes;

	width	= BufferInfo.dwSize.X;
	toprow	= BufferInfo.srWindow.Top;
	if ( len > ( width - 1 ) ) {
		len = width;
		svch = Text[len];
		Text[len] = '\0';
	}

	startcol = width - len - 1;


	//		Save contents of the screen region that the text will overwrite:

	ReadRegion.Top		= toprow;
	ReadRegion.Bottom	= toprow;
	ReadRegion.Left		= startcol;
	ReadRegion.Right	= width - 1;
	ReadConsoleOutput( ConsoleBufferHandle, TextbarSaveBuffer, TextbarSaveBufferSize, TextbarSaveBufferCoord, &ReadRegion );
	TextbarSaveCount	= len;
	TextbarSaveX		= startcol;
	TextbarSaveY		= toprow;


	//		Write the specified text to the screen:

	CursorPos.X = startcol;
	CursorPos.Y = toprow;
	SetConsoleCursorPosition( ConsoleBufferHandle, CursorPos );
	SetConsoleTextAttribute( ConsoleBufferHandle, TextBarColors );
	WriteConsole( ConsoleBufferHandle, Text, len, &CharsWritten, NULL );


	//		Restore the original cursor position and text color:

	CursorPos.X = BufferInfo.dwCursorPosition.X;
	CursorPos.Y = BufferInfo.dwCursorPosition.Y;
	SetConsoleCursorPosition( ConsoleBufferHandle, CursorPos );
	SetConsoleTextAttribute( ConsoleBufferHandle, OldColors );


	if ( svch )
		Text[len] = svch;

	return 0;
}



unsigned __int64 GetSystemTime()
{
	FILETIME FileTime = { 0, 0 };

	GetSystemTimeAsFileTime( &FileTime );

	return FileTime.dwLowDateTime | ( ( (unsigned __int64) FileTime.dwHighDateTime ) << 32 );
}



DLLExports int WINAPI _cmdcount( LPTSTR outString )
{
	if ( !outString )
		return -1;

	wsprintf( outString, L"%I64u", CommandCount );
	return 0;
}



int TicksToTime( unsigned __int64 Ticks, LPTSTR Buffer, BOOL Days )
{
	if ( !Buffer )
		return -1;


	unsigned int msec		= 0;
	unsigned int seconds	= 0;
	unsigned int minutes	= 0;
	unsigned int hours		= 0;

	if ( ( Ticks % 10000 ) >= 5000 )
		Ticks += 5000;

	Ticks /= 10000;
	msec = Ticks % 1000;
	Ticks /= 1000;
	seconds = Ticks % 60;
	Ticks /= 60;
	minutes = Ticks % 60;
	Ticks /= 60;

	if ( Days ) {
		hours = Ticks % 24;
		Ticks /= 24;
		wsprintf( Buffer, L"%I64u %02u:%02u:%02u.%03u", Ticks, hours, minutes, seconds, msec );
	}
	else
		wsprintf( Buffer, L"%I64u:%02u:%02u.%03u", Ticks, minutes, seconds, msec );

	return 0;
}



DLLExports int WINAPI _cmdticks( LPTSTR outString )
{
	if ( !outString )
		return -1;


	unsigned __int64 Duration = 0;

	if ( Running )
		Duration = GetSystemTime() - StartTime;
	else
		Duration = TimerDuration;

	wsprintf( outString, L"%I64u", Duration );
	return 0;
}



DLLExports int WINAPI _cmdtime( LPTSTR outString )
{
	if ( !outString )
		return -1;


	unsigned __int64 Duration = 0;


	if ( Running )
		Duration = GetSystemTime() - StartTime;
	else
		Duration = TimerDuration;

	TicksToTime( Duration, outString, FALSE );
	return 0;
}



DLLExports int WINAPI _cmdtotalticks( LPTSTR outString )
{
	if ( !outString )
		return -1;


	unsigned __int64 Duration = TotalDuration;
	unsigned int msec		= 0;
	unsigned int seconds	= 0;
	unsigned int minutes	= 0;


	if ( Running )
		Duration += ( GetSystemTime() - StartTime );

	if ( ResetTotalTime )
		Duration = 0;

	wsprintf( outString, L"%I64u", Duration );
	return 0;
}



DLLExports int WINAPI _cmdtotaltime( LPTSTR outString )
{
	if ( !outString )
		return -1;


	unsigned __int64 Duration = TotalDuration;


	if ( Running )
		Duration += ( GetSystemTime() - StartTime );

	if ( ResetTotalTime )
		Duration = 0;

	TicksToTime( Duration, outString, FALSE );
	return 0;
}



DLLExports int WINAPI _ticks( LPTSTR outString )
{
	if ( !outString )
		return -1;


	unsigned __int64 Ticks = GetSystemTime();

	wsprintf( outString, L"%I64u", Ticks );
	return 0;
}



DLLExports int WINAPI _lastcmd( LPTSTR outString )
{
	if ( !outString )
		return -1;


	if ( LastCommand )
		wcscpy_s( outString, MAX_ARG_LEN, LastCommand );
	else
		outString[0] = '\0';

	return 0;
}



DLLExports int WINAPI f_tickstotime( LPTSTR String )
{
	if ( !String )
		return -1;


	unsigned __int64 Ticks = 0;


	if ( String[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing ticks count\r\n" );
		return -1;
	}
	
	if ( ParseInt64( String, &Ticks ) ) {
		_wcsupr_s( String, MAX_ARG_LEN );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Bad number: \"%s\"\r\n", String );
		return -1;
	}
	
	TicksToTime( Ticks, String, FALSE );
	return 0;
}



DLLExports int WINAPI f_tickstotimed( LPTSTR String )
{
	if ( !String )
		return -1;


	unsigned __int64 Ticks = 0;


	if ( String[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing ticks count\r\n" );
		return -1;
	}
	
	if ( ParseInt64( String, &Ticks ) ) {
		_wcsupr_s( String, MAX_ARG_LEN );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Bad number: \"%s\"\r\n", String );
		return -1;
	}
	
	TicksToTime( Ticks, String, TRUE );
	return 0;
}



int StartTimer()
{
	if ( Running )
		return 0;

	StartTime = GetSystemTime();
	StopTime = 0;
	Running = TRUE;
	CommandCount++;
	return 0;
}



int StopTimer()
{
	if ( !Running )
		return 0;
	
	StopTime = GetSystemTime();
	TimerDuration = StopTime - StartTime;
	TotalDuration += TimerDuration;
	Running = FALSE;

	if ( ResetTotalTime ) {
		TotalDuration = 0;
		ResetTotalTime = FALSE;
	}

	return 0;
}



int DoSlowCommand()
{
	if ( !SlowCommand )
		return 0;


	wchar_t CmdBuf[MAX_ARG_LEN]		= L"";
	wcscpy_s( CmdBuf, MAX_ARG_LEN, SlowCommand );

	if ( CmdBuf[0] )
		Command( CmdBuf, 0 );

	return 0;
}



int SaveCommand( LPTSTR Command )
{
	if ( !Command )
		return -1;


	wchar_t buffer[MAX_ARG_LEN]		= L"";
	wchar_t *ch		= Command;
	BOOL quote		= FALSE;
	BOOL atsign		= FALSE;
	int j			= 0;


	if ( LastCommand ) {
		free( LastCommand );
		LastCommand = NULL;
	}


	while ( *ch && iswspace( *ch ) )
		ch++;

	while ( *ch ) {
		if ( *ch == QUOTE )
			quote = !quote;

		if ( ( *ch == '@' ) && !atsign && !j ) {
			ch++;
			atsign = TRUE;
			continue;
		}

		if ( iswspace( *ch ) && !quote )
			break;

		if ( j < ( MAX_ARG_LEN - 1 ) )
			buffer[j++] = *ch;

		ch++;
	}

	buffer[j] = '\0';


	if ( !j )
		return 0;


	LastCommand = (LPTSTR) malloc( ( j + 1 ) * sizeof( wchar_t ) );
	if ( !LastCommand )
		return 1;

	wcscpy_s( LastCommand, j + 1, buffer );
	return 0;
}



int ExpandVariablesQuietly( LPTSTR Buffer, int Flags = 1 )
{
	if ( !Buffer )
		return -1;


	HANDLE OldStderrHandle	= GetStdHandle( STD_ERROR_HANDLE );
	HANDLE OldStdoutHandle	= GetStdHandle( STD_OUTPUT_HANDLE );
	HANDLE NewStderrHandle	= CreateFile( L"nul", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL );
	int rv	= 0;

	if ( NewStderrHandle != INVALID_HANDLE_VALUE ) {
		SetStdHandle( STD_ERROR_HANDLE, NewStderrHandle );
		SetStdHandle( STD_OUTPUT_HANDLE, NewStderrHandle );
	}

	rv = ExpandVariables( Buffer, Flags );

	if ( NewStderrHandle != INVALID_HANDLE_VALUE ) {
		SetStdHandle( STD_ERROR_HANDLE, OldStderrHandle );
		SetStdHandle( STD_OUTPUT_HANDLE, OldStdoutHandle );
		CloseHandle( NewStderrHandle );
	}
	
	return rv;
}



BOOL AnythingThere( LPTSTR String )
{
	if ( !String )
		return FALSE;

	wchar_t *ch = String;

	while ( *ch ) {
		if ( !iswspace( *ch ) )
			return TRUE;

		ch++;
	}

	return FALSE;
}



DLLExports int WINAPI key( LPKEYINFO24 ki )
{
	int Key		= ki->nKey;

	if ( !Key && NewKeyHook ) {
		if ( ki->pszKey ) {
			if ( !wcscmp( ki->pszKey, L"Enter" ) )
				Key = CR;
			else if ( !wcscmp( ki->pszKey, L"Esc" ) )
				Key = ESC;
			else
				return 0;
		}
	}


	if ( !Key ) {
		//	_cputws( L"   * Key = 0; stop timer:\n" );

		BOOL WasRunning = Running;

		StopTimer();
		CloseTextBar();

		if ( ShowTimer && WasRunning ) {
			wchar_t Buffer[MAX_ARG_LEN] = L"";

			GetEnvironmentVariable( TimerFormatEnvVar, Buffer, MAX_ARG_LEN );

			if ( Buffer[0] == '\0' )
				wcscpy_s( Buffer, MAX_ARG_LEN, L" [%_cmdcount]  %_cmdtime " );

			ExpandVariablesQuietly( Buffer );

			if ( ShowTimer & SHOW_TEXTBAR_ALWAYS )
				ShowTextBar( Buffer );

			if ( ShowTimer & SHOW_STATUSBAR )
				if ( StatusBar( Buffer ) )
					if ( ShowTimer & SHOW_TEXTBAR_IF_STATUSBAR_FAILS )
						ShowTextBar( Buffer );
		}

		if ( WasRunning && SlowCommandTime && ( TimerDuration >= ( SlowCommandTime * TICKS_PER_SECOND ) ) )
			DoSlowCommand();
	}
	else if ( Key == CR ) {
		//	_cputws( L"   * Key = CR; start timer:\n" );

		if ( AnythingThere( ki->pszLine ) ) {
			StartTimer();
			SaveCommand( ki->pszLine );
		}

		CloseTextBar();
	}
	else if ( Key == ESC ) {
		//	_cputws( L"   * Key = ESC; close text bar:\n" );
		CloseTextBar();
	}

	return 0;
}



int GetColorNames( unsigned int Colors, LPTSTR outString )
{
	if ( !outString )
		return -1;

	wcscpy_s( outString, SHORT_BUF_LEN, ColorName[Colors & 0x0f] );
	wcscat_s( outString, SHORT_BUF_LEN, L" on " );
	wcscat_s( outString, SHORT_BUF_LEN, ColorName[ ( Colors & 0xf0 ) >> 4 ] );
	outString[0] = towupper( outString[0] );
	return 0;
}



int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion )
{
	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion )
		Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

	return 0;
}



DLLExports int WINAPI cmdtimer( LPTSTR inString )
{
	if ( !inString )
		return -1;


	wchar_t Colors[SHORT_BUF_LEN]	= L"";

	int SaveKey				= 0;
	BOOL Restore			= FALSE;
	BOOL PluginDefaults		= FALSE;
	BOOL ResetTotal			= FALSE;

	int NewShowTimer		= -1;
	int NewTextBarColors	= -1;
	int NewSlowCmdTime		= -1;
	LPTSTR NewSlowCommand	= NULL;

	int rv					= 0;
	int trv					= 0;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_ONE_ARG_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcscmp( arg[i], L"/?" ) )
				return ShowCmdHelp( CmdTimerHelpText, FALSE );
			else if ( !wcscmp( arg[i], L"/??" ) )
				return ShowCmdHelp( CmdTimerHelpText, TRUE );
			else if ( !_wcsnicmp( arg[i], L"/B", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				switch ( towupper( arg[i][j] ) ) {

				case 'B' :

					NewShowTimer	= SHOW_BOTH;
					break;

				case 'D' :

					NewShowTimer	= SHOW_DEFAULT;
					break;

				case 'S' :

					NewShowTimer	= SHOW_STATUSBAR;
					break;

				case 'T' :
				
					NewShowTimer	= SHOW_TEXTBAR_ALWAYS;
					break;

				case 'O' :

					NewShowTimer	= SHOW_NOTHING;
					break;

				default :

					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad suboption: \"%s\"\r\n", arg[i] );
					return 1;
				}
			}
			else if ( !_wcsnicmp( arg[i], L"/C", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				unsigned int Value = 0;

				if ( iswdigit( arg[i][j] ) ) {
					if ( ParseInt( arg[i] + j, &Value, NULL ) ) {
						_wcsupr_s( arg[i], MAX_ARG_LEN );
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Bad number: \"%s\"\r\n", arg[i] + j );
						return 1;
					}
				}
				else {
					int fg = 0xff, bg = 0xff;
					ParseColors( arg[i] + j, &fg, &bg );
					Value = fg | ( bg << 4 );
				}

				if ( ( Value > 0x00ff ) || ( ( ( Value & 0x00f0 ) >> 4 ) == ( Value & 0x000f ) ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad colors: \"%s\"\r\n", arg[i] + j );
					return 1;
				}

				NewTextBarColors = Value;
			}
			else if ( !_wcsicmp( arg[i], L"/D" ) )
				PluginDefaults = TRUE;
			else if ( !_wcsicmp( arg[i], L"/K" ) )
				CommandCount = 0;
			else if ( !_wcsnicmp( arg[i], L"/K", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				unsigned __int64 Value = 0;

				if ( ParseInt64( arg[i] + j, &Value ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad number: \"%s\"\r\n", arg[i] + j );
					return 1;
				}

				CommandCount = Value;
			}
			else if ( !_wcsicmp( arg[i], L"/R" ) )
				Restore = TRUE;
			else if ( !_wcsicmp( arg[i], L"/S" ) )
				SaveKey = 1;
			else if ( !_wcsicmp( arg[i], L"/S!" ) )
				SaveKey = 2;
			else if ( !_wcsnicmp( arg[i], L"/T", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( !_wcsicmp( arg[i] + j, L"Off" ) )
					NewSlowCmdTime = 0;
				else {
					unsigned int Value = 0;

					if ( ParseInt( arg[i] + j, &Value, NULL ) ) {
						_wcsupr_s( arg[i], MAX_ARG_LEN );
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Bad number: \"%s\"\r\n", arg[i] + j );
						return 1;
					}

					NewSlowCmdTime = Value;
				}
			}
			else if ( !_wcsicmp( arg[i], L"/Z" ) )
				ResetTotal = TRUE;
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option \"%s\"\r\n", arg[i] );
				return 1;
			}
			else
				NewSlowCommand = arg[i];

		}
	}


	//		Restore settings if desired :

	if ( PluginDefaults && Restore ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Can't combine /D and /R\r\n" );
		return 1;
	}


	if ( Restore ) {
		if ( trv = RestoreOptions() )
			Printf( L"Settings restored from the registry (%s\\%s).\r\n", RootKeyName( trv ), PluginRegKey );
		else
			Printf( L"Settings not found in the registry.\r\n" );
	}
	else if ( PluginDefaults ) {
		RestorePluginDefaults();
		Printf( L"Default plugin settings restored.\r\n" );
	}


	//		Make any requested changed :

	if ( NewShowTimer > -1 )
		ShowTimer = NewShowTimer;

	if ( NewTextBarColors > -1 )
		TextBarColors = NewTextBarColors;

	if ( NewSlowCmdTime > -1 )
		SlowCommandTime = NewSlowCmdTime;

	if ( NewSlowCommand ) {
		if ( wcscmp( NewSlowCommand, L"-" ) )
			SetSlowCommand( NewSlowCommand );
		else
			SetSlowCommand( NULL );
	}

	if ( ResetTotal )
		ResetTotalTime = TRUE;


	//		Display the current settings :

	Printf( L"\r\n%s v%u.%u.%u settings:\r\n\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );

	Printf( L"   Show timer . . . . . . . . /B : %s\r\n", ShowModeStr( ShowTimer ) );

	GetColorNames( TextBarColors, Colors );
	Printf( L"   Timer text colors  . . . . /C : %s\r\n", Colors );

	if ( SlowCommandTime )
		Printf( L"   Slow commands take . . . . /T : %u second%s or more\r\n", SlowCommandTime, Plural( SlowCommandTime ) );
	else
		QPuts( L"   Slow commands take . . . . /T : 0 seconds (disabled)\r\n" );
	
	Printf( L"   Slow-Command . . . . . . . %s\r\n", SlowCommand ? SlowCommand : L"Not defined" );


	//		Save the current settings if desired :

	if ( SaveKey ) {
		if ( trv = SaveOptions( ( SaveKey > 1 ) ? HKEY_LOCAL_MACHINE : HKEY_CURRENT_USER ) ) {
			CrLf();
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error saving settings to the registry (%s\\%s)\r\n", RootKeyName( SaveKey ), PluginRegKey );
			rv = 2;
		}
		else
			Printf( L"\r\nSettings saved to the registry (%s\\%s).\r\n", RootKeyName( SaveKey ), PluginRegKey );
	}

	return rv;
}



LPTSTR UsageText( const wchar_t *inString )
{
	if ( !inString )
		return NULL;


	wchar_t *ch = wcschr( (wchar_t*) inString, '\n' );


	if ( !ch )
		return NULL;

	while ( *ch == '\n' )
		ch++;

	return ch;
}



DLLExports int WINAPI Help( LPTSTR inString )
{
	if ( !_wcsicmp( inString, L"cmdtimer" ) ) {
		MessageBox( GetForegroundWindow(), CmdTimerHelpText, HelpWindowTitle, MB_OK );
		return 1;
	}
	else if ( !_wcsicmp( inString, L"_cmdcount" ) ) {
		MessageBox( GetForegroundWindow(), _CmdCountHelpText, HelpWindowTitle, MB_OK );
		return 1;
	}
	else if ( !_wcsicmp( inString, L"_cmdticks" ) ) {
		MessageBox( GetForegroundWindow(), _CmdTicksHelpText, HelpWindowTitle, MB_OK );
		return 1;
	}
	else if ( !_wcsicmp( inString, L"_cmdtime" ) ) {
		MessageBox( GetForegroundWindow(), _CmdTimeHelpText, HelpWindowTitle, MB_OK );
		return 1;
	}
	else if ( !_wcsicmp( inString, L"_cmdtotalticks" ) ) {
		MessageBox( GetForegroundWindow(), _CmdTotalTicksHelpText, HelpWindowTitle, MB_OK );
		return 1;
	}
	else if ( !_wcsicmp( inString, L"_cmdtotaltime" ) ) {
		MessageBox( GetForegroundWindow(), _CmdTotalTimeHelpText, HelpWindowTitle, MB_OK );
		return 1;
	}
	else if ( !_wcsicmp( inString, L"_lastcmd" ) ) {
		MessageBox( GetForegroundWindow(), _LastCmdHelpText, HelpWindowTitle, MB_OK );
		return 1;
	}
	else if ( !_wcsicmp( inString, L"_ticks" ) ) {
		MessageBox( GetForegroundWindow(), _TicksHelpText, HelpWindowTitle, MB_OK );
		return 1;
	}
	else if ( !_wcsicmp( inString, L"@tickstotime" ) ) {
		MessageBox( GetForegroundWindow(), f_TicksToTimeHelpText, HelpWindowTitle, MB_OK );
		return 1;
	}
	else if ( !_wcsicmp( inString, L"@tickstotimed" ) ) {
		MessageBox( GetForegroundWindow(), f_TicksToTimeDHelpText, HelpWindowTitle, MB_OK );
		return 1;
	}
	else
		return 0;
}



DLLExports int WINAPI Usage( LPTSTR inString, LPTSTR outString )
{
	if ( !_wcsicmp( inString, L"cmdtimer" ) ) {
		wcscpy_s( outString, SHORT_BUF_LEN, UsageText( CmdTimerHelpText ) );
		return 1;
	}
	else
		return 0;
}



/*

	1.0.0.0		2021-07-08
				First release.

	1.0.1.0		2021-12-21
				Fix for an issue calling TakeCommandIPC() on 32-bit TCC.

	1.0.2.0		2021-12-22
				No longer link to TakeCommandIPC() dynamically.  That hasn't been
				necessary for a long long time.

	1.0.2.1		2023-10-12
				Updated the plugin's web address.

	1.0.2.2		2023-10-16
				Tweaked ShowCmdHelp() to include VER_PATCH.

	1.0.2.3		2023-11-16
				"Plugin not loaded" message is not display in transient/pipe shells;
				now includes the plugin's name.

*/
